/* stack.c - implementation of the stack data structure. */
/* Copyright (C) 2014 Robin Miyagi
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <dstralg.h>

#include <config.h>
#include <stddef.h>
#include <stdlib.h>

struct dstralg_stack_struct
{
  dstralg_destruct_T destruct;
  const void **top;
  size_t items;
};
typedef struct dstralg_stack_struct *dstralg_stack_T;

/* FORWARD DECLARATIONS ........................................... */
dstralg_stack_T dstralg_stack_new (dstralg_destruct_T destruct);
void dstralg_stack_delete (dstralg_stack_T stack);
int dstralg_stack_push (dstralg_stack_T stack, const void *data);
void *dstralg_stack_pop (dstralg_stack_T stack);

/* IMPLEMENTATIONS ................................................ */
dstralg_stack_T
dstralg_stack_new (dstralg_destruct_T destruct)
{
  dstralg_stack_T stack;
  
  stack = calloc (1, sizeof (struct dstralg_stack_struct));
  stack->destruct = destruct;
  return stack;
}
void
dstralg_stack_delete (dstralg_stack_T stack)
{
  const void **bottom;
  size_t i;
  
  bottom = stack->top - stack->items;
  for (i = 0; i < stack->items; ++i)
    stack->destruct ((void *) bottom[i]);
  free (bottom);
  free (stack);
}
int
dstralg_stack_push (dstralg_stack_T stack, const void *data)
{
  const void **bottom;

  bottom = stack->top - stack->items;
  ++(stack->items);
  bottom = realloc (bottom, stack->items * sizeof (void *));
  if (bottom == NULL)
    return -1;
  stack->top = bottom + stack->items;
  *(stack->top - 1) = data;
  return stack->items;
}
void
*dstralg_stack_pop (dstralg_stack_T stack)
{
  const void **bottom;
  const void *data;

  if (stack->items == 0)
    return NULL;
  data = *(stack->top - 1);
  bottom = stack->top - stack->items;
  --(stack->items);
  if (stack->items == 0)
    {
      free (bottom);
      stack->top = NULL;
    }
  else
    {
      bottom = realloc (bottom, stack->items * sizeof (void *));
      if (bottom == NULL)
	return (void *) -1;
      else
	stack->top = bottom + stack->items;
    }
  return (void *) data;
}
